#include"k3water.h"

#include<string.h>
#include<stdio.h>

struct k3Water *k3WaterCreate(float w, float h, size_t subdivisions, struct k3Mat *mat) {
	size_t edgesubdiv = 2 + subdivisions;
	
	size_t vertices = edgesubdiv * edgesubdiv;
	
	struct k3Water *ret = malloc(sizeof(*ret));
	
	ret->w = w;
	ret->h = h;
	
	for(int f = 0; f < k3_WATER_FRAMES; f++) {
		ret->frames[f] = malloc(sizeof(vec3) * vertices);
	}
	
	ret->curFrame = 0;
	
	ret->dx = w / (edgesubdiv - 1);
	ret->dy = h / (edgesubdiv - 1);
	
	ret->ptsPerEdge = edgesubdiv;
	
	for(int z = 0; z < edgesubdiv; z++) {
		for(int x = 0; x < edgesubdiv; x++) {
			ret->frames[0][z * edgesubdiv + x][0] = x * w / (edgesubdiv - 1) - w / 2;
			ret->frames[0][z * edgesubdiv + x][1] = 0;
			ret->frames[0][z * edgesubdiv + x][2] = z * h / (edgesubdiv - 1) - h / 2;
		}
	}
	memcpy(ret->frames[1], ret->frames[0], sizeof(vec3) * vertices);
	memcpy(ret->frames[2], ret->frames[0], sizeof(vec3) * vertices);
	
	size_t indices = (1 + subdivisions) * (1 + subdivisions) * 6;
	uint16_t *inds = malloc(sizeof(*inds) * indices);
	
	for(int z = 0; z <= subdivisions; z++) {
		for(int x = 0; x <= subdivisions; x++) {
			inds[6 * (z * (subdivisions + 1) + x) + 0] = (z * edgesubdiv + x) + 0;
			inds[6 * (z * (subdivisions + 1) + x) + 1] = (z * edgesubdiv + x) + edgesubdiv;
			inds[6 * (z * (subdivisions + 1) + x) + 2] = (z * edgesubdiv + x) + edgesubdiv + 1;
			inds[6 * (z * (subdivisions + 1) + x) + 3] = (z * edgesubdiv + x) + 0;
			inds[6 * (z * (subdivisions + 1) + x) + 4] = (z * edgesubdiv + x) + edgesubdiv + 1;
			inds[6 * (z * (subdivisions + 1) + x) + 5] = (z * edgesubdiv + x) + 1;
		}
	}
	
	uint8_t *norms = alloca(ret->ptsPerEdge * ret->ptsPerEdge * sizeof(*norms) * 3);
	
	ret->mdl = k3MdlCreate(vertices, indices, 0, ret->frames[0], norms, NULL, NULL, NULL, NULL, inds, NULL, NULL);
	
	k3MdlAddMesh(ret->mdl, mat, 0, indices);
	
	free(inds);
	
	return ret;
}

static void k3WaterSimSubStep(struct k3Water *this, float dt, float C, float mu) {
	int prevPrevFrame = (this->curFrame + (k3_WATER_FRAMES - 1)) % k3_WATER_FRAMES;
	int prevFrame = this->curFrame;
	this->curFrame = (this->curFrame + 1) % k3_WATER_FRAMES;
	
	vec3 *prevPrev = this->frames[prevPrevFrame];
	vec3 *prev = this->frames[prevFrame];
	vec3 *cur = this->frames[this->curFrame];
	
	for(int z = 1; z < this->ptsPerEdge - 1; z++) {
		for(int x = 1; x < this->ptsPerEdge - 1; x++) {
			cur[z * this->ptsPerEdge + x][1] =
				+ 4 * prev[z * this->ptsPerEdge + x][1] / (2 + mu * dt)
				- prevPrev[z * this->ptsPerEdge + x][1] * (2 - mu * dt) / (2 + mu * dt)
				+ 2 * C * C * dt * dt * (
					(prev[z * this->ptsPerEdge + x + 1][1] - 2 * prev[z * this->ptsPerEdge + x][1] + prev[z * this->ptsPerEdge + x - 1][1]) / (this->dx * this->dx) +
					(prev[(z + 1) * this->ptsPerEdge + x][1] - 2 * prev[z * this->ptsPerEdge + x][1] + prev[(z - 1) * this->ptsPerEdge + x][1]) / (this->dy * this->dy)
				);
		}
	}
	
	for(int x = 1; x < this->ptsPerEdge - 1; x++) {
		cur[0 * this->ptsPerEdge + x][1] = 0;
		cur[(this->ptsPerEdge - 1) * this->ptsPerEdge + x][1] = 0;
	}
	
	for(int z = 0; z < this->ptsPerEdge; z++) {
		cur[z * this->ptsPerEdge + 0][1] = 0;
		cur[z * this->ptsPerEdge + this->ptsPerEdge - 1][1] = 0;
	}
}

void k3WaterSimulate(struct k3Water *this, float dt, float C, float mu, int substeps) {
	for(int i = 0; i < substeps; i++) {
		k3WaterSimSubStep(this, dt / substeps, C, mu);
	}
}

void k3WaterUpdate(struct k3Water *this) {
	k3MdlUpdatePos(this->mdl, this->frames[this->curFrame]);
	
	typedef int8_t Nrm[3];
	
	Nrm *nrm = alloca(this->ptsPerEdge * this->ptsPerEdge * sizeof(*nrm));
	
	for(size_t i = 0; i < this->ptsPerEdge; i++) {
		nrm[i * this->ptsPerEdge + 0][0] = 0;
		nrm[i * this->ptsPerEdge + 0][1] = 1;
		nrm[i * this->ptsPerEdge + 0][2] = 0;
		
		nrm[i * this->ptsPerEdge + this->ptsPerEdge - 1][0] = 0;
		nrm[i * this->ptsPerEdge + this->ptsPerEdge - 1][1] = 1;
		nrm[i * this->ptsPerEdge + this->ptsPerEdge - 1][2] = 0;
		
		nrm[0 * this->ptsPerEdge + i][0] = 0;
		nrm[0 * this->ptsPerEdge + i][1] = 1;
		nrm[0 * this->ptsPerEdge + i][2] = 0;
		
		nrm[(this->ptsPerEdge - 1) * this->ptsPerEdge + i][0] = 0;
		nrm[(this->ptsPerEdge - 1) * this->ptsPerEdge + i][1] = 1;
		nrm[(this->ptsPerEdge - 1) * this->ptsPerEdge + i][2] = 0;
	}
	
	for(size_t z = 1; z < this->ptsPerEdge - 1; z++) {
		for(size_t x = 1; x < this->ptsPerEdge - 1; x++) {
			vec3 n = {
				(this->frames[this->curFrame][z * this->ptsPerEdge + x + 1][1] - this->frames[this->curFrame][z * this->ptsPerEdge + x - 1][1]) / 2,
				1,
				(this->frames[this->curFrame][(z + 1) * this->ptsPerEdge + x][1] - this->frames[this->curFrame][(z - 1) * this->ptsPerEdge + x][1]) / 2,
			};
			
			glm_vec3_normalize(n);
			
			nrm[z * this->ptsPerEdge + x][0] = n[0] * 127;
			nrm[z * this->ptsPerEdge + x][1] = n[1] * 127;
			nrm[z * this->ptsPerEdge + x][2] = n[2] * 127;
		}
	}
	
	k3MdlUpdateNrm(this->mdl, nrm);
}

void k3WaterDisturb(struct k3Water *this, float x, float y, float radius, float powr) {
	int i = (x + this->w / 2) * this->ptsPerEdge / this->w;
	int j = (y + this->h / 2) * this->ptsPerEdge / this->h;
	
	this->frames[this->curFrame][j * this->ptsPerEdge + i][1] += -powr;
	this->frames[(this->curFrame + 2) % k3_WATER_FRAMES][j * this->ptsPerEdge + i][1] += -powr;
}
